from .description import Description
from .encoder import Encoder
from .jsonencoder import JSON_Encoder
from .multiencoder import MultiEncoder
from .questionnaire import Questionnaire
from .scoring import Scoring
from .label import Label
from .generate_json_encoding import generate_json_encoding
from .interval_label import Interval_Label
from .quartile_label import Quartile_Label

__all__ = [
    "Description",
    "Encoder",
    "JSON_Encoder",
    "MultiEncoder",
    "Questionnaire",
    "Scoring",
    "Label",
    "generate_json_encoding",
    "Interval_Label",
    "Quartile_Label",
]


import pandas as pd
import warnings
from ..util import generate_name

class Encoder(object):
    name_generator = generate_name("encoder")

    def __init__(self, encoding=None, *, template=None, inverse=False, default=None, neutral=None, dtype=None, name=None):
        if encoding is not None:
            self.encoding = encoding
            self.target = self.encoding.keys()
            self.default = default
            self.neutral = neutral
            if dtype is None:
                self.dtype = str
            else:
                self.dtype = dtype
        
        elif template is not None:
            # Default and neutral will be overwritten later if they are specified in kwargs.
            self.encoding = template.encoding
            self.target = self.encoding.keys()
            self.default = default if default is not None else template.default
            self.neutral = neutral if neutral is not None else template.neutral
            self.dtype = dtype if dtype is not None else template.dtype
            
        else:
            raise ValueError("Expected argument 'encoding'")

        if inverse:
            self.inverse()
        
        # TODO: If template is used, we can use other name instead.
        if name is None:
            self.name = next(Encoder.name_generator)
        
        else:
            self.name = name
    
    def inverse(self):
        encode_item = sorted(self.encoding.items(), key=lambda x:x[1], reverse=True)
        target = [i for i, j in encode_item]
        values = [j for i, j in encode_item]
        self.encoding = dict(zip(target, sorted(values)))
        
    
    def transform(self, data, *, columns=None, ignore_list=None, return_rule=False, mode='any'):
        result = data.copy()

        if isinstance(data, pd.Series):
            result.replace(self.encoding, inplace=True)
            if self.default is not None:
                result.fillna(self.default, inplace=True)
            return result
        
        elif isinstance(data, pd.DataFrame):
            encode_rule = pd.Series(index=data.columns, dtype=str)

            if ignore_list is None:
                ignore_list = []

            # TODO: Rewrite this section
            for i in result.columns:
                if i in ignore_list:
                    continue

                if columns is not None:
                    if i in columns:
                        result[i] = result[i].replace(self.encoding)
                        if self.default is not None:
                            result[i] = result[i].fillna(self.default)
                        encode_rule[i] = self.name

                else:
                    unique_values = result[i].value_counts().index
                    if mode == "strict":
                        if len(set(unique_values) ^ set(self.encoding.keys())) == 0:
                            result[i] = result[i].replace(self.encoding)
                            if self.default is not None:
                                result[i] = result[i].fillna(self.default)
                            encode_rule[i] = self.name
                    elif mode == "any":
                        if len(set(unique_values) - set(self.encoding.keys())) == 0:
                            result[i] = result[i].replace(self.encoding)
                            if self.default is not None:
                                result[i] = result[i].fillna(self.default)
                            encode_rule[i] = self.name
                    else:
                        raise ValueError("rule argument can only be strict or any")
            
            if return_rule:
                return (result, encode_rule)

            else:
                return result
        else:
            raise TypeError(f"Expected pandas Series or DataFrame, got {type(data)} instead")
    
    def count_neutral(self, data, **kwargs):
        if self.neutral is not None:
            if isinstance(data, pd.Series):
                total = (self.transform(data) == self.neutral).astype(int)
                return total

            elif isinstance(data, pd.DataFrame):
                return_flag = False
                if 'return_rule' in kwargs.keys() and kwargs['return_rule']:
                    return_flag = True
                
                else:
                    kwargs['return_rule'] = True

                df_encoded, rule = self.transform(data, **kwargs)
                total = (df_encoded[rule.dropna().index] == self.neutral).sum(axis=1)

                if return_flag:
                    return (total, rule)
                
                else:
                    return total
            
            else:
                raise TypeError(f"Expected pandas Series or DataFrame, got {type(data)} instead")

        else:
            warnings.warn(f"Encoder {self.name} does not have neutral argument specified, count_neutral will return None")
            return None



import itertools

from matplotlib import pyplot as plt
import seaborn as sns

def generate_name(prefix):
  for i in itertools.count():
    yield f"{prefix}-{i}"

def plotter(f):
  def plotter_function(*args, figsize=(12, 12), title='Big title', **kwargs):
    plt.figure(figsize=figsize, tight_layout=True)
    f(*args, **kwargs)
    figure = plt.gcf()
    figure.suptitle(title, fontsize=16, y=1.05)
  return plotter_function
  
@plotter
def plot_each_col(data, 
                  col_list, 
                  plot_type, 
                  n_col=2, 
                  x=None,
                  top=10, 
                  **kwargs):
  '''
  Plot a subplot of specified type on each selected column. 

  Arguments:
  data: Input DataFrame
  col_list: The columns to be plotted.
  n_col: Number of subplots on each row.
  plot_type: Graph type.
  x: The column for x-axis, used for graphs type like line and trend graph.
  top: For "top" plot_type. If positive, get the top most frequent values, else get the least frequent values.
  '''
  n_row = len(col_list) // n_col + 1
  for i, col in enumerate(col_list):
    ax = plt.subplot(n_row, n_col, i + 1)
    if plot_type == "hist":
      sns.histplot(data=data, x=col, multiple="stack", **kwargs)
    
    elif plot_type == "bar":
      sns.barplot(data=data, x=col, **kwargs)

    elif plot_type == "count":
      sns.countplot(data=data, x=col, **kwargs)

    elif plot_type == "box":
      sns.boxplot(data=data, x=col, **kwargs)
    
    elif plot_type == "line":
      if x:
        sns.lineplot(data=data, x=x, y=col, ax=ax, **kwargs)

      else:
        sns.lineplot(data=data, x=data.index, y=col, ax=ax, **kwargs)
    
    elif plot_type == "trend":
      plot_trend(data=data, x=x, y=col, ax=ax, **kwargs)

    elif plot_type == "top":
      temp = data[col].value_counts()
      if top > 0:
        sns.barplot(x=temp.index[0:top], y=temp[0:top])
      else:
        sns.barplot(x=temp.index[-1:top:-1], y=temp[-1:top:-1])

    else:
      raise ValueError(f"Invalid plot_type argument: {plot_type}")

    ax.set_title(f"Distribution of {col}")

def create_moving_average(data, average=7, min_periods=1):
  return data.rolling(average, min_periods=min_periods).mean()

def plot_trend(data, y, x=None, ax=None, 
               date_index=True, date_index_name=None, 
               moving_average=None, min_periods=1,
               label=None, ax_format=None):
  '''
  Plot a line graph on the trend on a new or existing ax object.

  Arguments:
  data: A pandas DataFrame. Do not pass a pandas Series
  y: Column name for plotting y-axis
  ax: If None, plot on a new ax object.
  date_index: If passed, the x-axis will be formatted nicely for a date_index
  date_index_name: The index level name holding the date values.
  moving_average: If integer is passed, will create a moving average on y-value.
  min_periods: min_periods for moving_average.
  label: Name for legend
  ax_format: Function that takes an ax for formatting.
  '''
  if date_index:
    if date_index_name is None:
      raise ValueError("Must pass in date_index_name")
    
    if x is not None:
      raise ValueError("Cannot pass x argument when setting date_index to True")
    
    x = data.index.get_level_values(date_index_name).map(lambda x: dt.datetime.strptime(str(x), "%Y-%m-%d %H:%M:%S"))

  if label is None:
    label = y

  if moving_average is not None:
    data = create_moving_average(data[[y]], average=moving_average, min_periods=min_periods)
  
  if ax is None:
    ax = sns.lineplot(data=data, x=x, y=y, label=label)

    if date_index:
      ax_format(ax)

  else:
    ax = sns.lineplot(data=data, x=x, y=y, ax=ax, label=label)

  return ax



import pandas as pd
from tickcounter.questionnaire import Label
from ..util import generate_name

class Interval_Label(Label):
    def __init__(self, label_rule, name=None):
        self.label_rule = label_rule
        super().__init__(self.generate_label_function(self.label_rule), name)
    
    def generate_label_function(self, label_rule):
        def label(data, score_col):
            # Only support numerical pandas series, will support multiple score_col in the future
            label_ss = pd.Series(index=data.index, dtype=str)
            for label, interval in label_rule.items():
                label_ss[data[score_col].between(*interval)] = label
            return label_ss
        return label


import pandas as pd
from tickcounter.questionnaire import Label
from ..util import generate_name

class Quartile_Label(Label):
    def __init__(self, q, labels, name=None):
        self.q = q
        self.labels = labels
        super().__init__(self.generate_label_function(self.q, self.labels), name)
    
    def generate_label_function(self, q, labels):
        def label(data, score_col):
            label_ss = pd.qcut(data[score_col], q=q, labels=labels)
            return label_ss
        return label


import pandas as pd
from ..util import generate_name

from tickcounter.questionnaire import Encoder

class MultiEncoder(object):
  def __init__(self, encoding_rule):
    if isinstance(encoding_rule, Encoder):
      self.rules = {
        encoding_rule.name: encoding_rule
      }
    
    elif isinstance(encoding_rule, list):
      if isinstance(encoding_rule[0], Encoder):
        self.rules = dict()
        for i in encoding_rule:
          self.rules[i.name] = i

      else:
        pass
        # Need to convert the dictionary to encoders, and give default name
    
    else:
      raise ValueError(f"Expected list of encoder or dictionary objects, got {type(encoding_rule)} instead")
    
  def transform(self, data,*, rule_map=None, columns=None, ignore_list=None, return_rule=False, mode="any"):
    result = data.copy()
    encode_rule = None
    if isinstance(data, pd.DataFrame):
      encode_rule = pd.Series(dtype=str, index=data.columns)
      if rule_map is None:
        for i in result.columns:
          if ignore_list is not None and i in ignore_list:
            continue

          else:
            unique_values = result[i].value_counts().index
            for rule in self.rules.values():
              if mode == "strict":
                if len(set(unique_values) ^ set(rule.target)) == 0:
                  result[i] = rule.transform(result[i])
                  encode_rule[i] = rule.name
                  break
              elif mode == "any":
                if len(set(unique_values) - set(rule.target)) == 0:
                  result[i] = rule.transform(result[i])
                  encode_rule[i] = rule.name
                  break
              else:
                raise ValueError("rule argument can only be strict or any")
      else:
        # Check for correct format for rule_map
        # Transform according to the rules
        pass

    elif isinstance(data, pd.Series):
      encode_rule = pd.Series(dtype=str, index=[data.name])
      unique_values = result.value_counts().index
      for rule in self.rules.values():
        if mode == "strict":
          if len(set(unique_values) ^ set(rule.target)) == 0:
            result = rule.transform(result)
            encode_rule[data.name] = rule.name
            break
        elif mode == "any":
          if len(set(unique_values) - set(rule.target)) == 0:
            result = rule.transform(result)
            encode_rule[data.name] = rule.name
            break
        else:
          raise ValueError("rule argument can only be strict or any")

    else: 
      raise TypeError(f"Expected pandas Series or DataFrame, got {type(data)} instead")
    
    if return_rule:
      return (result, encode_rule)
    
    else:
      return result
    
  def count_neutral(self, data, **kwargs):
    # Might need to refactor this
    return_flag = False
    if 'return_rule' in kwargs.keys() and kwargs['return_rule']:
        return_flag = True
    else:
        kwargs['return_rule'] = True

    df_encoded, rule = self.transform(data, **kwargs)
    total = None
    if isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
      for col, encoder in rule.dropna().iteritems():
        # Need to rewrite this. We transform the thing twice to get the count of neutral! 
        ss_tally = self.rules[encoder].count_neutral(data[col] if isinstance(data, pd.DataFrame) else data)
        # If encoder does not have neutral, it will return None
        if ss_tally is not None:
          if total is None:
            total = pd.DataFrame([ss_tally]).T
          
          else:
            total = pd.concat([total, ss_tally], axis=1)
        else:
          continue
      
      # None will result if there is no neutral specified
      if total is not None:
        total = total.sum(axis=1)
        total.rename("Neutral count", inplace=True)
    
    if return_flag:
      return (total, rule)
    
    else:
      return total


import pandas as pd
from ..util import generate_name

class Label(object):
  name_generator = generate_name("label")

  def __init__(self, label_function, name=None):
    self.label_function = label_function
    self.name = name
    if self.name is None:
      self.name = next(Label.name_generator)
  
  def label(self, data, score_col, **kwargs):
    return self.label_function(data, score_col, **kwargs)


import pandas as pd
import warnings
from ..util import generate_name

class Scoring(object):
    name_generator = generate_name("scoring")
    def __init__(self, labeling, encoding=None,*, name=None, columns=None):
        self.encoding = encoding
        self.name = name
        self.labeling = labeling if isinstance(labeling, list) else [labeling] # Used for categorization

        if self.encoding is not None:
            self.columns = []

            for i in encoding.values():
                self.columns.extend(i)
        
        elif (self.encoding is None) and (columns is not None):
            self.columns = columns
        
        elif (self.encoding is not None) and (columns is not None):
            raise ValueError("Can only pass either encoding or columns argument.")

        else:
            raise ValueError("Expected either encoding or columns argument.")

        if self.name is None:
            self.name = next(Scoring.name_generator)

    def transform(self, data, **kwargs):
        if self.encoding is not None:
            df = data.copy()
            if isinstance(self.encoding, dict):
                for encoder, columns in self.encoding.items():
                    df[columns] = encoder.transform(df[columns])
                
                return df
            
            elif isinstance(self.encoding, MultiEncoder):
                if 'return_rule' in kwargs and kwargs['return_rule']:
                    df, rule = self.encoding.transform(df, **kwargs)
                    return df, rule
                
                else:
                    df = self.encoding.transform(df, **kwargs)
                    return df
            
            else:
                raise ValueError(f"Invalid encoder type {type(self.encoding)}")
        else:
            warnings.warn(f"Scoring object {self.name} has no encoding, original DataFrame is returned")
            return data

    def score(self, data):
        df = self.transform(data)
        score_ss = df[self.columns].sum(axis=1)
        score_ss.rename(f"{self.name} score", inplace=True)
        return score_ss

    def label(self, data, score_col):
        label_df = None
        
        for i in self.labeling:
            label_ss = i.label(data, score_col)
            label_ss.rename(f"{self.name} - Label {i.name}", inplace=True)
        
            if label_df is None:
                label_df = pd.DataFrame([label_ss]).T
        
            else:
                label_df = pd.concat([label_df, label_ss], axis=1)
        
        return label_df
    
    @property
    def score_col(self):
        return f"{self.name} score"
    
    @property
    def label_col(self):
        return [f"{self.name} - Label {i.name}" for i in self.labeling]


import pandas as pd
from itertools import combinations
from scipy.stats import ttest_ind, chisquare
from ..util import generate_name, plot_each_col

class Questionnaire(object):
    def __init__(self, data, scoring, descrip=None):
        self.data = data.copy() # Original data
        self.descrip = descrip # Used for plotting and analysis, takes a Description object
        self._cached = {
            "transform": None, # Will only contain transformed columns
            "score": None,
            "label": None,
            "data": self.data,
        }
        self.scoring = scoring if isinstance(scoring, list) else [scoring] # Used for calculating score
    
    def transform(self):
        if self._cached['transform'] is not None:
            return self._cached['transform']

        else:
            data = None
            # Will change this to support MultiEncoder in the future.
            for i in self.scoring:
                cur_data = i.transform(self.data[i.columns])
                if data is None:
                    data = cur_data
                else:
                    data = pd.concat([data, cur_data], axis=1)
            self._cached['transform'] = data
            return data

    def score(self):
        if self._cached['score'] is not None:
            return self._cached['score']
        
        else:
            result_df = None
            for i in self.scoring:
                score_ss = i.score(self.data)
                
                if result_df is None:
                    result_df = pd.DataFrame([score_ss]).T
                
                else:
                    result_df = pd.concat([result_df, score_ss], axis=1)
            
            self._cached['score'] = result_df
            self._cached['data'] = pd.concat([self._cached['data'], self._cached['score']], axis=1)
            return result_df

    def label(self):
        if self._cached['label'] is not None:
            return self._cached['label']

        else:
            data = None
            if self._cached['score'] is not None:
                data = self._cached['data']
            
            else:
                self.score()
                data = self._cached['data']

            label_df = None
            for scoring in self.scoring:
                cur_label_df = scoring.label(data, scoring.score_col)
                
                if label_df is None:
                    label_df = cur_label_df 
                
                else:
                    label_df = pd.concat([label_df, cur_label_df], axis=1)
            self._cached['label'] = label_df
            self._cached['data'] = pd.concat([self._cached['data'], label_df], axis=1)
            return label_df

    def _plot(self, columns, kind, transformed, **kwargs):
        data = None
        if self._cached['label'] is not None:
            data = self._cached['data']
        
        else:
            self.label()
            data = self._cached['data']
        if transformed:
            data[self.item_col] = self._cached['transform']
        plot_each_col(data, col_list = columns, plot_type=kind, **kwargs)

    def hist_label(self, *, transformed=True, separated=False, **kwargs):
        self._plot(columns = self.label_col, kind='hist', transformed=transformed, **kwargs)

    def hist_item(self, *, transformed=True, separated=False, **kwargs):
        self._plot(columns = self.item_col, kind='hist', transformed=transformed, **kwargs)
    
    def hist_score(self, *, transformed=True, separated=False, **kwargs):
        self._plot(columns = self.score_col, kind='hist', transformed=transformed, **kwargs)

    def boxplot_score(self, *, transformed=True, separated=False, **kwargs):
        self._plot(columns = self.score_col, kind='box', transformed=transformed, **kwargs)
    
    def boxplot_item(self, *, transformed=True, separated=False, **kwargs):
        self._plot(columns = self.item_col, kind='box', transformed=transformed, **kwargs)
    
    def count_label(self, *, transformed=True, separated=False, **kwargs):
        self._plot(columns = self.item_col, kind='count', transformed=transformed, **kwargs)
    
    def diff(self, col, transformed=True):
        if col not in self._cached['data'].columns:
            self.label()
        
        if transformed:
            df_group = self._cached['data'].copy()
            df_group[self.item_col] = self.transform()
        
        else:
            df_group = self._cached['data']

        df_group = df_group.groupby(col)[self.item_col].mean().T
        result = pd.DataFrame(index=df_group.index)
        for i in combinations(df_group.columns, 2):
            result[f"{i[0]} - {i[1]}"] = df_group[i[0]] - df_group[i[1]]
        return result
    
    def crosstab(self, index, col):
        # Should be a label paired with an info columns
        self.label()
        return pd.crosstab(self._cached["data"][index], self._cached["data"][col])
    
    def t_test(self, item, info_col, **kwargs):
        # Compare the mean of the scoring or item between two different groups
        df = self._cached['data']
        if info_col not in self._cached['data'].columns:
            self.label()
            df = self._cached['data']
        test_result = dict()
        for i in combinations(df[info_col].value_counts().index, r=2):
            test_result[f"{i[0]} vs {i[1]}"] = ttest_ind(a = df[df[info_col] == i[0]][item],
                                                         b = df[df[info_col] == i[1]][item],
                                                         **kwargs)
        return test_result
    
    def chi_squared(self, info_col):
        self.label()
        df = self._cached['data']
        test_result = dict()
        for i in self.scoring:
            # test for chi_squared
            pass
        return

    def cluster(self, scoring):
        # Use KMeans clustering to cluster the response to something
        pass

    def drop(self, idx):
        self.data.drop(idx, inplace=True)
        self.reset_cache()

    def reset_cache(self):
        self._cached = {
            "transform": None,
            "score": None,
            "label": None,
            "data": self.data,
        }
    
    @property
    def score_col(self):
        if self._cached['score'] is not None:
            return self._cached['score'].columns
        
        else:
            self.score()
            return self._cached['score'].columns
    
    @property
    def label_col(self):
        if self._cached['label'] is not None:
            return self._cached['label'].columns
        
        else:
            self.label()
            return self._cached['label'].columns
    
    @property
    def item_col(self):
        if self._cached['transform'] is not None:
            return self._cached['transform'].columns
        
        else:
            self.transform()
            return self._cached['transform'].columns
    
    def __getitem__(self, col):
        try:
            return self._cached['data'][col]

        except KeyError:
            self.label()
            return self._cached['data'][col]

    


class Description(object):
  "Do we want to change this to dictionary object?"
  "Refer to https://stackoverflow.com/questions/3387691/how-to-perfectly-override-a-dict/39375731#39375731"
  def __init__(self, descrip):
    self.descrip = descrip


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tickcounter', 'tickcounter.questionnaire', 'tickcounter.util']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.2.2', 'pandas>=1.3.5', 'scipy>=1.4.1', 'seaborn>=0.11.2']

setup_kwargs = {
    'name': 'tickcounter',
    'version': '0.1.1',
    'description': 'A library for processing questionnaire data',
    'long_description': 'Work in progress',
    'author': 'Ong Eng Kheng',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1',
}


setup(**setup_kwargs)



import json

def generate_json_encoding(data, *, neutral=False, default=False, filename="encoding.json", dtype=str):
    # Use this one https://pypi.org/project/keyboard/, for implementing undo option and also skip
    # Might need to use sampling if the data is very large
    encoding = dict()
    for col in data.columns:
        if input(f"Encode {col}? (y/n) ") == 'n':
            continue 
        else:
            cur_dict = dict()
            mapping_dict = dict()
            for value in data[col].value_counts().index:
                mapping_dict[value] = dtype(prompt_encoding(col, value))
            cur_dict['encoding'] = mapping_dict
            if neutral:
                neutral_value = prompt_neutral(col, value)
                if neutral_value != '':
                    cur_dict['neutral'] = dtype(neutral_value)
            if default:
                default_value = prompt_default(col, value)
                if default_value != '':
                    cur_dict['default'] = dtype(default_value)
            encoding[col] = cur_dict
    with open(filename, 'w') as f:
        json.dump(encoding, f, indent=4)
    return json.dumps(encoding, indent=4)


def prompt_encoding(col_name, value):
    inp = input(f"{col_name} - encode '{value}' as? ")
    return inp

def prompt_neutral(col_name, value):
    inp = input(f"{col_name} - neutral value? ")
    return inp

def prompt_default(col_name, value):
    inp = input(f"{col_name} - default value? ")
    return inp
